// Provides a [color.Color] type for dealing with premultiplied linear RGBA colours.
package lrgba

import (
	"image/color"
	"math"

	"smariot.com/color/internal/helper"
)

// Color is a pre-multiplied linear RGBA [color.Color].
type Color struct {
	R, G, B, A float64
}

func sqr(a float64) float64 {
	return a * a
}

// DistanceSqr returns the maximum possible euclidean distance squared between two colours,
// accounting for the possible backgrounds they might be composited over.
func DistanceSqr(a, b Color) float64 {
	dR := a.R - b.R
	dG := a.G - b.G
	dB := a.B - b.B
	dA := a.A - b.A
	return max(sqr(dR), sqr(dR+dA)) + max(sqr(dG), sqr(dG+dA)) + max(sqr(dB), sqr(dB+dA))
}

// Distance returns the maximum possible euclidean distance between two colours,
// accounting for the possible backgrounds they might be composited over.
//
// If you just want to compare relative distances, use [DistanceSqr] instead.
func Distance(a, b Color) float64 {
	return math.Sqrt(DistanceSqr(a, b))
}

// RGBA converts to premultiplied RGBA, implementing the [color.Color] interface.
func (c Color) RGBA() (r, g, b, a uint32) {
	return helper.NLRGBAtoRGBA(c.NLRGBA())
}

// NRGBA converts to non-premultiplied RGBA.
func (c Color) NRGBA() (r, g, b, a uint32) {
	return helper.NLRGBAtoNRGBA(c.NLRGBA())
}

// NLRGBA converts to non-premultiplied linear RGBA.
func (c Color) NLRGBA() (r, g, b, a float64) {
	if c.A <= 0 {
		return 0, 0, 0, 0
	}

	return c.R / c.A, c.G / c.A, c.B / c.A, c.A
}

// NXYZA converts to non-premultiplied XYZ+Alpha.
func (c Color) NXYZA() (x, y, z, a float64) {
	if c.A <= 0 {
		return 0, 0, 0, 0
	}

	x, y, z = helper.LRGBtoXYZ(c.R/c.A, c.G/c.A, c.B/c.A)
	return x, y, z, c.A
}

// NOkLabA converts to non-premultiplied OkLab+Alpha.
func (c Color) NOkLabA() (lightness, chromaA, chromaB, a float64) {
	if c.A <= 0 {
		return 0, 0, 0, 0
	}

	lightness, chromaA, chromaB = helper.LMStoOkLab(helper.LRGBtoLMS(c.R/c.A, c.G/c.A, c.B/c.A))
	return lightness, chromaA, chromaB, c.A
}

// Convert converts an arbitrary colour type to a premultiplied linear RGBA [Color].
func Convert(c color.Color) Color {
	if c, ok := c.(Color); ok {
		return c
	}

	r, g, b, a := helper.ColorToNLRGBA(c)
	return Color{r * a, g * a, b * a, a}
}

// A [color.Model] for converting arbitrary colours to a premultiplied linear RGBA [Color].
//
// Wraps the [Convert] function, returning a [color.Color] interface rather than the [Color] type.
var Model = helper.Model(Convert)

// Type assertion.
var _ color.Color = Color{}
